#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['quasar']

package_data = \
{'': ['*']}

install_requires = \
['numpy >= 1.18.4', 'scipy >= 1.3.3', 'sortedcontainers==2.1.0']

extras_require = \
{'cirq': ['cirq == 0.5.0'],
 'ionq': ['requests >= 2.22.0'],
 'pyquil': ['pyquil==2.16.0'],
 'qiskit': ['qiskit-aer >= 0.5.1']}

setup(name='qcware-quasar',
      version='1.0.0',
      description='Build and execute quantum circuits on various backends.',
      author='Rob Parrish',
      author_email='rob.parrish@qcware.com',
      url='http://github.com/qcware/quasar',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )
